# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used by debusine wait tasks."""

import datetime as dt
from enum import StrEnum
from typing import override

from debusine.tasks.models import (
    BaseDynamicTaskData,
    BaseTaskData,
    LookupSingle,
)


class DelayData(BaseTaskData):
    """In-memory task data for the Delay task."""

    delay_until: dt.datetime


class ExternalDebsignData(BaseTaskData):
    """In-memory task data for the ExternalDebsign task."""

    unsigned: LookupSingle


class ExternalDebsignDynamicData(BaseDynamicTaskData):
    """Dynamic data for the ExternalDebsign task."""

    unsigned_id: int

    @override
    def get_input_artifacts_ids(self) -> list[int]:
        return [self.unsigned_id]


class DenyAction(StrEnum):
    """The action to take when denying a Confirm task."""

    FAIL = "fail"
    ABORT = "abort"
    ABORT_WORKFLOW = "abort-workflow"


class ConfirmData(BaseTaskData):
    """In-memory task data for the Confirm task."""

    auto_confirm_if_no_failure: bool = False
    deny_action: DenyAction = DenyAction.ABORT_WORKFLOW
    confirm_label: str = "Continue the workflow"
    deny_label: str = "Abort the workflow"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for debusine."""

from debusine.web.views.autopkgtest import AutopkgtestViewWorkRequestPlugin
from debusine.web.views.debdiff import DebDiffViewArtifactPlugin
from debusine.web.views.debian_pipeline import (
    DebianPipelineViewWorkRequestPlugin,
)
from debusine.web.views.lintian import LintianViewWorkRequestPlugin
from debusine.web.views.qa import QAViewWorkRequestPlugin
from debusine.web.views.views import HomepageView

__all__ = [
    "AutopkgtestViewWorkRequestPlugin",
    "DebDiffViewArtifactPlugin",
    "DebianPipelineViewWorkRequestPlugin",
    "HomepageView",
    "LintianViewWorkRequestPlugin",
    "QAViewWorkRequestPlugin",
]



# Generated by Django 4.2.28 on 2026-04-09 06:39

import django.db.models.constraints
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import debusine.artifacts.models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0050_alter_token_token_type_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='CollectionRelation',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('source_category', models.CharField(max_length=255)),
                ('target_category', models.CharField(max_length=255)),
                (
                    'type',
                    models.CharField(
                        choices=[
                            ('suite-forked-from', 'Forked from'),
                            ('suite-based-on', 'Based on'),
                            ('suite-requires', 'Requires'),
                            ('suite-targeting', 'Targeting'),
                            ('suite-default-qa-results', 'Default QA Results'),
                        ],
                        max_length=32,
                    ),
                ),
                (
                    'position',
                    models.PositiveIntegerField(blank=True, null=True),
                ),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(blank=True, null=True)),
                (
                    'created_by_user',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name='created_collection_relations',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    'source',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name='relations',
                        to='db.collection',
                    ),
                ),
                (
                    'target',
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name='targeted_by',
                        to='db.collection',
                    ),
                ),
                (
                    'updated_by_user',
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name='updated_collection_relations',
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                'base_manager_name': 'objects',
            },
        ),
        migrations.AddConstraint(
            model_name='collectionrelation',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(
                        ('position__isnull', False), ('type', 'suite-requires')
                    ),
                    models.Q(
                        ('position__isnull', True),
                        (
                            'type__in',
                            [
                                'suite-forked-from',
                                'suite-based-on',
                                'suite-targeting',
                                'suite-default-qa-results',
                            ],
                        ),
                    ),
                    _connector='OR',
                ),
                name='db_collectionrelation_position',
            ),
        ),
        migrations.AddConstraint(
            model_name='collectionrelation',
            constraint=models.UniqueConstraint(
                deferrable=django.db.models.constraints.Deferrable['DEFERRED'],
                fields=('source', 'type', 'position'),
                name='db_collectionrelation_suite_requires_no_duplicate_position',
            ),
        ),
        migrations.AddConstraint(
            model_name='collectionrelation',
            constraint=models.UniqueConstraint(
                condition=models.Q(
                    ('type', 'suite-forked-from'),
                    ('type', 'suite-based-on'),
                    ('type', 'suite-targeting'),
                    ('type', 'suite-default-qa-results'),
                    _connector='OR',
                ),
                fields=('source', 'type'),
                name='db_collectionrelation_cardinality_checks_max_1',
            ),
        ),
        migrations.AddConstraint(
            model_name='collectionrelation',
            constraint=models.UniqueConstraint(
                condition=models.Q(('type', 'suite-requires')),
                fields=('source', 'type', 'target'),
                name='db_collectionrelation_suite_requires_no_duplicate_target',
            ),
        ),
        migrations.AddConstraint(
            model_name='collectionrelation',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(
                        (
                            'source_category',
                            debusine.artifacts.models.CollectionCategory[
                                'SUITE'
                            ],
                        ),
                        (
                            'target_category',
                            debusine.artifacts.models.CollectionCategory[
                                'SUITE'
                            ],
                        ),
                        ('type', 'suite-forked-from'),
                    ),
                    models.Q(
                        (
                            'source_category',
                            debusine.artifacts.models.CollectionCategory[
                                'SUITE'
                            ],
                        ),
                        (
                            'target_category',
                            debusine.artifacts.models.CollectionCategory[
                                'SUITE'
                            ],
                        ),
                        ('type', 'suite-based-on'),
                    ),
                    models.Q(
                        (
                            'source_category',
                            debusine.artifacts.models.CollectionCategory[
                                'SUITE'
                            ],
                        ),
                        (
                            'target_category',
                            debusine.artifacts.models.CollectionCategory[
                                'SUITE'
                            ],
                        ),
                        ('type', 'suite-requires'),
                    ),
                    models.Q(
                        (
                            'source_category',
                            debusine.artifacts.models.CollectionCategory[
                                'SUITE'
                            ],
                        ),
                        (
                            'target_category',
                            debusine.artifacts.models.CollectionCategory[
                                'SUITE'
                            ],
                        ),
                        ('type', 'suite-targeting'),
                    ),
                    models.Q(
                        (
                            'source_category',
                            debusine.artifacts.models.CollectionCategory[
                                'SUITE'
                            ],
                        ),
                        (
                            'target_category',
                            debusine.artifacts.models.CollectionCategory[
                                'QA_RESULTS'
                            ],
                        ),
                        ('type', 'suite-default-qa-results'),
                    ),
                    _connector='OR',
                ),
                name='db_collectionrelation_categories_correct',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the reverse_dependencies_autopkgtest workflow."""

from collections.abc import Iterable, Mapping, Sequence
from typing import Any, ClassVar, override

from debusine.artifacts.local_artifact import Upload
from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebianBinaryPackage,
    DebianSourcePackage,
    TaskTypes,
)
from debusine.client.models import LookupChildType
from debusine.db.models import Artifact, Collection, WorkRequest
from debusine.server.workflows import (
    ReverseDependenciesAutopkgtestWorkflow,
    WorkflowValidationError,
)
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import (
    BaseWorkflowData,
    RegressionTrackingUpdateQAResults,
    ReverseDependenciesAutopkgtestWorkflowData,
    ReverseDependenciesAutopkgtestWorkflowDynamicData,
    SbuildWorkflowData,
)
from debusine.server.workflows.reverse_dependencies_autopkgtest import (
    NoBinaryNames,
    _BinaryPackage,
    _SourcePackage,
)
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestCase,
)
from debusine.tasks import TaskConfigError
from debusine.tasks.models import (
    ActionSkipIfLookupResultChanged,
    ActionUpdateCollectionWithArtifacts,
    AutopkgtestNeedsInternet,
    BackendType,
    BaseDynamicTaskData,
    ExtraExternalRepository,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    LookupSingle,
    SbuildData,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry


class ReverseDependenciesAutopkgtestWorkflowTests(
    WorkflowTestCase[ReverseDependenciesAutopkgtestWorkflow]
):
    """Unit tests for :py:class:`ReverseDependenciesAutopkgtestWorkflow`."""

    SAMPLE_TASK_DATA = {
        "binary_artifacts": ["internal@collections/name:build-amd64"],
        "qa_suite": "sid@debian:suite",
        "vendor": "debian",
        "codename": "sid",
    }

    sid: ClassVar[Collection]
    source_artifact: ClassVar[Artifact]
    binary_artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.sid = cls.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )
        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )
        cls.binary_artifact = (
            cls.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello", package="hello"
            )
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data["source_artifact"] = self.source_artifact.pk
        self.internal_collection_artifacts = [
            (self.binary_artifact, {"name": "build-amd64"}),
        ]

    def create_source_package(
        self,
        *,
        name: str,
        version: str,
        dsc_fields: dict[str, Any] | None = None,
        add_to_collection: Collection | None = None,
    ) -> Artifact:
        """Create a `debian:source-package` artifact."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data=DebianSourcePackage(
                name=name,
                version=version,
                type="dpkg",
                dsc_fields={
                    "Package": name,
                    "Version": version,
                    **(dsc_fields or {}),
                },
            ),
        )
        if add_to_collection is not None:
            add_to_collection.manager.add_artifact(
                artifact,
                user=self.playground.get_default_user(),
                variables={"component": "main", "section": "devel"},
            )
        return artifact

    def create_binary_package(
        self,
        *,
        source_package: Artifact,
        name: str | None = None,
        version: str | None = None,
        architecture: str = "all",
        deb_fields: dict[str, Any] | None = None,
        add_to_collection: Collection | None = None,
    ) -> Artifact:
        """Create a `debian:binary-package` artifact."""
        source_package_data = source_package.create_data()
        assert isinstance(source_package_data, DebianSourcePackage)
        if name is None:
            name = source_package_data.name
        if version is None:
            version = source_package_data.version
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data=DebianBinaryPackage(
                srcpkg_name=source_package_data.name,
                srcpkg_version=source_package_data.version,
                deb_fields={
                    "Package": name,
                    "Version": version,
                    "Architecture": architecture,
                    **(deb_fields or {}),
                },
                deb_control_files=[],
            ),
        )
        if add_to_collection is not None:
            add_to_collection.manager.add_artifact(
                artifact,
                user=self.playground.get_default_user(),
                variables={
                    "component": "main",
                    "section": "devel",
                    "priority": "optional",
                },
            )
        return artifact

    def test_create_orchestrator(self) -> None:
        """A ReverseDependenciesAutopkgtestWorkflow can be instantiated."""
        self.assertEqual(
            self.task.data.source_artifact, self.source_artifact.pk
        )
        self.assertEqual(
            self.task.data.binary_artifacts,
            LookupMultiple(self.SAMPLE_TASK_DATA["binary_artifacts"]),
        )
        self.assertEqual(
            self.task.data.qa_suite, self.SAMPLE_TASK_DATA["qa_suite"]
        )
        self.assertEqual(self.task.data.vendor, "debian")
        self.assertEqual(self.task.data.codename, "sid")
        self.assertEqual(self.task.data.backend, BackendType.UNSHARE)

    def test_create_orchestrator_explicit_backend(self) -> None:
        """A ReverseDependenciesAutopkgtestWorkflow can take a backend."""
        self.task_data["backend"] = BackendType.INCUS_LXC
        self.assertEqual(self.task.data.backend, BackendType.INCUS_LXC)

    def test_validate_input(self) -> None:
        """validate_input passes a valid case."""
        self.task.validate_input()

    def test_validate_input_bad_qa_suite(self) -> None:
        """validate_input raises errors in looking up qa_suite."""
        self.validate = False
        self.task_data["qa_suite"] = "nonexistent@debian:suite"
        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:suite' does not exist or is hidden",
        ):
            self.task.validate_input()

    def test_get_binary_names_promise(self) -> None:
        """Get binary names from a promise."""
        # Delete the automatically-added artifact from the collection,
        # to exercise the promise code
        assert self.work_request.internal_collection is not None
        self.work_request.internal_collection.child_artifacts.clear()

        self.playground.create_bare_data_item(
            self.work_request.internal_collection,
            "build-amd64",
            category=BareDataCategory.PROMISE,
            data={
                "promise_work_request_id": self.work_request.id + 1,
                "promise_workflow_id": self.work_request.id,
                "promise_category": ArtifactCategory.UPLOAD,
                "binary_names": ["hello", "hello-dev"],
            },
        )

        self.assertEqual(self.task.get_binary_names(), {"hello", "hello-dev"})

    def test_get_binary_names_binary_package(self) -> None:
        """Get binary names from a `debian:binary-package` artifact."""
        temp_dir = self.create_temporary_directory()
        binary_package = self.playground.create_artifact_from_local(
            self.playground.create_binary_package(temp_dir, name="single")
        )
        self.task_data["binary_artifacts"] = [binary_package.id]
        self.assertEqual(self.task.get_binary_names(), {"single"})

    def test_get_binary_names_upload(self) -> None:
        """Get binary names from a `debian:upload` artifact."""
        temp_dir = self.create_temporary_directory()
        changes_path = temp_dir / "hello_1.0_amd64.changes"
        self.write_changes_file(
            changes_path, [], binaries=["hello", "hello-dev"]
        )
        upload = self.playground.create_artifact_from_local(
            Upload.create(changes_file=changes_path)
        )
        self.task_data["binary_artifacts"] = [upload.id]
        self.assertEqual(self.task.get_binary_names(), {"hello", "hello-dev"})

    def test_get_binary_names_bad_artifact_category(self) -> None:
        """Cannot get binary names from a non-binary artifact category."""
        temp_dir = self.create_temporary_directory()
        source_package = self.playground.create_artifact_from_local(
            self.playground.create_source_package(temp_dir)
        )
        self.task_data["binary_artifacts"] = [source_package.id]
        with self.assertRaisesRegex(
            NoBinaryNames,
            f"Artifact of category {ArtifactCategory.SOURCE_PACKAGE} has no "
            f"binary packages",
        ):
            self.task.get_binary_names()

    def test_get_reverse_dependencies_bad_source_artifact_category(
        self,
    ) -> None:
        """The source artifact must be a `debian:source-package`."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.TEST
        )
        self.task_data["source_artifact"] = artifact.id

        with self.assertRaisesRegex(
            TaskConfigError,
            r"^source_artifact: unexpected artifact category: 'debusine:test'. "
            r"Valid categories: \['debian:source-package', 'debian:upload'\]$",
        ):
            self.task

    def assert_reverse_dependencies(
        self,
        workflow: ReverseDependenciesAutopkgtestWorkflow,
        lookups: Mapping[LookupSingle, Iterable[LookupSingle]],
    ) -> None:
        """Assert that reverse-dependencies of a workflow are correct."""
        expected = []
        for source_lookup, binary_lookups in lookups.items():
            source = workflow.work_request.lookup_single(
                source_lookup,
                default_category=CollectionCategory.SUITE,
                expect_type=LookupChildType.ARTIFACT,
            ).artifact
            source_data = source.create_data()
            assert isinstance(source_data, DebianSourcePackage)
            expected_source = _SourcePackage(
                qa_suite=workflow.data.qa_suite,
                name=source_data.name,
                version=source_data.version,
            )
            expected_binaries = []
            for binary_lookup in binary_lookups:
                binary = workflow.work_request.lookup_single(
                    binary_lookup,
                    default_category=CollectionCategory.SUITE,
                    expect_type=LookupChildType.ARTIFACT,
                ).artifact
                binary_data = binary.create_data()
                assert isinstance(binary_data, DebianBinaryPackage)
                expected_binaries.append(
                    _BinaryPackage(
                        qa_suite=workflow.data.qa_suite,
                        name=binary_data.deb_fields["Package"],
                        version=binary_data.deb_fields["Version"],
                        architecture=binary_data.deb_fields["Architecture"],
                    )
                )
            expected.append(
                (
                    expected_source,
                    sorted(
                        expected_binaries,
                        key=lambda binary: (
                            binary.name,
                            binary.version,
                            binary.architecture,
                        ),
                    ),
                )
            )

        rdeps = workflow.get_reverse_dependencies()

        self.assertEqual(
            rdeps,
            sorted(expected, key=lambda rdep: (rdep[0].name, rdep[0].version)),
        )
        # The lookups are all cached.
        for source_package, binary_packages in rdeps:
            with self.assertNumQueries(0):
                workflow.work_request.lookup_single(
                    source_package.to_lookup(),
                    expect_type=LookupChildType.ARTIFACT,
                )
            for binary_package in binary_packages:
                with self.assertNumQueries(0):
                    workflow.work_request.lookup_single(
                        binary_package.to_lookup(),
                        expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
                    )

    def test_get_reverse_dependencies(self) -> None:
        """Get reverse dependencies of a source package and its binaries."""
        trixie = self.playground.create_collection(
            name="trixie", category=CollectionCategory.SUITE
        )

        pre_depends = self.create_source_package(
            name="pre-depends",
            version="1.0",
            dsc_fields={"Testsuite": "autopkgtest"},
            add_to_collection=self.sid,
        )
        self.create_binary_package(
            source_package=pre_depends,
            name="python3-pre-depends",
            version="1:1.0",
            deb_fields={"Pre-Depends": "python3-urllib3:any"},
            add_to_collection=self.sid,
        )

        depends = self.create_source_package(
            name="depends",
            version="2.32.3+dfsg-1",
            dsc_fields={
                "Testsuite": "autopkgtest-pkg-python, autopkgtest-pkg-pybuild"
            },
            add_to_collection=self.sid,
        )
        self.create_binary_package(
            source_package=depends,
            name="python3-depends",
            deb_fields={"Depends": "python3-urllib3 (>= 1.21.1), python3:any"},
            add_to_collection=self.sid,
        )
        self.create_binary_package(
            source_package=depends,
            name="depends-doc",
            deb_fields={"Depends": "libjs-sphinxdoc"},
            add_to_collection=self.sid,
        )

        similar_depends = self.create_source_package(
            name="similar-depends",
            version="1",
            dsc_fields={"Testsuite": "autopkgtest-pkg-pybuild"},
            add_to_collection=self.sid,
        )
        self.create_binary_package(
            source_package=similar_depends,
            name="python3-similar-depends",
            deb_fields={"Depends": "python3-urllib3-plugin"},
            add_to_collection=self.sid,
        )

        for version, add_to_collection in (("1", trixie), ("2", self.sid)):
            trigger = self.create_source_package(
                name="trigger",
                version=version,
                dsc_fields={
                    "Testsuite": "autopkgtest",
                    "Testsuite-Triggers": "python3-urllib3",
                },
                add_to_collection=add_to_collection,
            )
            self.create_binary_package(
                source_package=trigger, add_to_collection=add_to_collection
            )

        previous_version = self.create_source_package(
            name="python-urllib3",
            version="2.0.7-1",
            dsc_fields={
                "Testsuite": "autopkgtest",
                "Testsuite-Triggers": "python3-urllib3",
            },
            add_to_collection=self.sid,
        )
        self.create_binary_package(
            source_package=previous_version, name="python3-urllib3"
        )

        unrelated = self.create_source_package(
            name="unrelated",
            version="1",
            dsc_fields={"Testsuite": "autopkgtest"},
            add_to_collection=self.sid,
        )
        self.create_binary_package(
            source_package=unrelated, add_to_collection=add_to_collection
        )

        untested = self.create_source_package(
            name="untested", version="1", add_to_collection=self.sid
        )
        self.create_binary_package(
            source_package=untested,
            deb_fields={"Depends": "python3-urllib3"},
            add_to_collection=self.sid,
        )

        unsuitable_testsuite = self.create_source_package(
            name="unsuitable-testsuite",
            version="1",
            dsc_fields={
                "Testsuite": "autopkgtest-unsuitable",
                "Testsuite-Triggers": "python3-urllib3",
            },
            add_to_collection=self.sid,
        )
        self.create_binary_package(
            source_package=unsuitable_testsuite, add_to_collection=self.sid
        )

        source_package = self.create_source_package(
            name="python-urllib3", version="2.0.7-2"
        )
        binary_package = self.create_binary_package(
            source_package=source_package, name="python3-urllib3"
        )
        workflow = self.task_create(
            source_artifact=source_package.id,
            binary_artifacts=[binary_package.id],
            qa_suite=self.sid.id,
        )

        self.assert_reverse_dependencies(
            workflow,
            {
                "sid/source-version:pre-depends_1.0": (
                    "sid/binary-version:python3-pre-depends_1:1.0_all",
                ),
                "sid/source-version:depends_2.32.3+dfsg-1": (
                    "sid/binary-version:python3-depends_2.32.3+dfsg-1_all",
                    "sid/binary-version:depends-doc_2.32.3+dfsg-1_all",
                ),
                "sid/source-version:trigger_2": (
                    "sid/binary-version:trigger_2_all",
                ),
            },
        )

        workflow.data.packages_denylist = ["depends"]
        self.assert_reverse_dependencies(
            workflow,
            {
                "sid/source-version:pre-depends_1.0": (
                    "sid/binary-version:python3-pre-depends_1:1.0_all",
                ),
                "sid/source-version:trigger_2": (
                    "sid/binary-version:trigger_2_all",
                ),
            },
        )

        workflow.data.packages_denylist = []
        workflow.data.packages_allowlist = ["depends"]
        self.assert_reverse_dependencies(
            workflow,
            {
                "sid/source-version:depends_2.32.3+dfsg-1": (
                    "sid/binary-version:python3-depends_2.32.3+dfsg-1_all",
                    "sid/binary-version:depends-doc_2.32.3+dfsg-1_all",
                ),
            },
        )

        upload_artifacts = self.playground.create_upload_artifacts(
            src_name="python-urllib3",
            version="2.0.7-2",
            binaries=[("python3-urllib3", "all")],
        )
        workflow = self.task_create(
            source_artifact=upload_artifacts.upload.id,
            binary_artifacts=[upload_artifacts.binaries[0].id],
            qa_suite=self.sid.id,
        )
        self.assert_reverse_dependencies(
            workflow,
            {
                "sid/source-version:pre-depends_1.0": (
                    "sid/binary-version:python3-pre-depends_1:1.0_all",
                ),
                "sid/source-version:depends_2.32.3+dfsg-1": (
                    "sid/binary-version:python3-depends_2.32.3+dfsg-1_all",
                    "sid/binary-version:depends-doc_2.32.3+dfsg-1_all",
                ),
                "sid/source-version:trigger_2": (
                    "sid/binary-version:trigger_2_all",
                ),
            },
        )

    def orchestrate(
        self,
        source_artifact: Artifact,
        architectures: Sequence[str],
        *,
        extra_data: dict[str, Any] | None = None,
        priority_base: int = 0,
        parent: WorkRequest | None = None,
        pipeline_task_name: str = "examplepipeline",
    ) -> WorkRequest:
        """Create a workflow and call ``orchestrate_workflow``."""

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Pipeline workflow that runs sbuild and rdep-autopkgtest."""

            TASK_NAME = pipeline_task_name

            @override
            def populate(self_) -> None:
                sbuild = self_.work_request.create_child_workflow(
                    task_name="sbuild",
                    task_data=SbuildWorkflowData(
                        input=SbuildInput(source_artifact=source_artifact.id),
                        target_distribution="debian:sid",
                        architectures=["all", *architectures],
                    ),
                )
                for architecture in ("all", *architectures):
                    child = sbuild.create_child_worker(
                        task_name="sbuild",
                        task_data=SbuildData(
                            input=SbuildInput(
                                source_artifact=source_artifact.id
                            ),
                            build_architecture=architecture,
                            environment="debian/match:codename=sid",
                        ),
                    )
                    self_.provides_artifact(
                        child,
                        ArtifactCategory.UPLOAD,
                        f"build-{architecture}",
                        data={
                            "binary_names": ["hello"],
                            "architecture": architecture,
                        },
                    )

                rdep_autopkgtest = self_.work_request.create_child_workflow(
                    task_name="reverse_dependencies_autopkgtest",
                    task_data=ReverseDependenciesAutopkgtestWorkflowData(
                        source_artifact=source_artifact.id,
                        binary_artifacts=LookupMultiple.model_validate(
                            [
                                f"internal@collections/"
                                f"name:build-{architecture}"
                                for architecture in ("all", *architectures)
                            ]
                        ),
                        qa_suite=f"sid@{CollectionCategory.SUITE}",
                        vendor="debian",
                        codename="sid",
                        **(extra_data or {}),
                    ),
                )
                self.playground.advance_work_request(
                    rdep_autopkgtest, mark_pending=True
                )
                self_.orchestrate_child(rdep_autopkgtest)

        root = self.playground.create_workflow(
            task_name=pipeline_task_name, parent=parent
        )
        root.priority_base = priority_base
        root.save()
        self.assertTrue(orchestrate_workflow(root))

        return root.children.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="reverse_dependencies_autopkgtest",
        )

    @preserve_db_task_registry()
    def test_populate(self) -> None:
        """The workflow populates child work requests."""
        depends = self.create_source_package(
            name="depends",
            version="1.0",
            dsc_fields={"Testsuite": "autopkgtest"},
            add_to_collection=self.sid,
        )
        self.create_binary_package(
            source_package=depends,
            deb_fields={"Depends": "hello"},
            add_to_collection=self.sid,
        )
        trigger = self.create_source_package(
            name="trigger",
            version="1.0",
            dsc_fields={
                "Testsuite": "autopkgtest",
                "Testsuite-Triggers": "hello",
            },
            add_to_collection=self.sid,
        )
        self.create_binary_package(
            source_package=trigger, add_to_collection=self.sid
        )
        source_artifact = self.source_artifact
        architectures = ("amd64", "i386")

        rdep_autopkgtest = self.orchestrate(
            source_artifact,
            architectures,
            extra_data={
                "extra_repositories": [
                    ExtraExternalRepository(
                        url="http://example.com/",
                        suite="bookworm",
                        components=["main"],
                    )
                ]
            },
        )

        children = list(
            WorkRequest.objects.filter(parent=rdep_autopkgtest).order_by(
                "task_data__prefix"
            )
        )
        for child, source in zip(children, ("depends_1.0", "trigger_1.0")):
            self.assertEqual(child.status, WorkRequest.Statuses.RUNNING)
            self.assertEqual(child.task_type, TaskTypes.WORKFLOW)
            self.assertEqual(child.task_name, "autopkgtest")
            self.assertEqual(
                child.task_data,
                {
                    "prefix": f"{source}|",
                    "reference_prefix": f"{source}|",
                    "source_artifact": (
                        f"sid@{CollectionCategory.SUITE}/"
                        f"source-version:{source}"
                    ),
                    "binary_artifacts": [
                        f"sid@{CollectionCategory.SUITE}/"
                        f"binary-version:{source}_all"
                    ],
                    "context_artifacts": [
                        f"internal@collections/name:build-{architecture}"
                        for architecture in ("all", *architectures)
                    ],
                    "qa_suite": f"sid@{CollectionCategory.SUITE}",
                    "reference_qa_results": None,
                    "enable_regression_tracking": False,
                    "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                    "vendor": "debian",
                    "codename": "sid",
                    "backend": BackendType.UNSHARE,
                    "architectures": [],
                    "arch_all_build_architecture": "amd64",
                    "debug_level": 0,
                    "extra_repositories": [
                        {
                            "url": "http://example.com/",
                            "suite": "bookworm",
                            "components": ["main"],
                        }
                    ],
                },
            )
            self.assertEqual(child.priority_base, -5)

            grandchild = WorkRequest.objects.get(parent=child)
            self.assertEqual(grandchild.status, WorkRequest.Statuses.BLOCKED)
            self.assertEqual(grandchild.task_type, TaskTypes.WORKER)
            self.assertEqual(grandchild.task_name, "autopkgtest")
            self.assertEqual(
                grandchild.task_data,
                {
                    "input": {
                        "source_artifact": (
                            f"sid@{CollectionCategory.SUITE}/"
                            f"source-version:{source}"
                        ),
                        "binary_artifacts": [
                            f"sid@{CollectionCategory.SUITE}/"
                            f"binary-version:{source}_all"
                        ],
                        "context_artifacts": [
                            f"internal@collections/name:build-{architecture}"
                            for architecture in ("all", "amd64")
                        ],
                    },
                    "build_architecture": "amd64",
                    "environment": "debian/match:codename=sid",
                    "backend": BackendType.UNSHARE,
                    "include_tests": [],
                    "exclude_tests": [],
                    "debug_level": 0,
                    "extra_environment": {},
                    "extra_repositories": [
                        {
                            "url": "http://example.com/",
                            "suite": "bookworm",
                            "components": ["main"],
                        }
                    ],
                    "pin_packages": {},
                    "needs_internet": AutopkgtestNeedsInternet.RUN,
                    "fail_on": {},
                    "timeout": None,
                },
            )
            self.assertEqual(grandchild.priority_base, -5)
            self.assert_work_request_event_reactions(
                grandchild,
                on_success=[
                    ActionUpdateCollectionWithArtifacts(
                        collection="internal@collections",
                        name_template=f"{source}|autopkgtest-amd64",
                        artifact_filters={
                            "category": ArtifactCategory.AUTOPKGTEST
                        },
                    )
                ],
            )
            self.assertQuerySetEqual(
                grandchild.dependencies.all(),
                list(
                    WorkRequest.objects.filter(
                        task_type=TaskTypes.WORKER,
                        task_name="sbuild",
                        task_data__build_architecture__in={"all", "amd64"},
                    )
                ),
            )
            self.assertEqual(
                grandchild.workflow_data_json,
                {
                    "display_name": "autopkgtest amd64",
                    "step": "autopkgtest-amd64",
                },
            )

        # Population is idempotent.
        self.get_workflow(rdep_autopkgtest).populate()
        children = list(WorkRequest.objects.filter(parent=rdep_autopkgtest))
        self.assertEqual(len(children), 2)

    @preserve_db_task_registry()
    def test_populate_passes_regression_tracking_parameters(self) -> None:
        """The workflow passes on parameters related to regression tracking."""
        self.playground.create_collection(
            name="sid", category=CollectionCategory.QA_RESULTS
        )
        depends = self.create_source_package(
            name="depends",
            version="1.0",
            dsc_fields={"Testsuite": "autopkgtest"},
            add_to_collection=self.sid,
        )
        binary_artifact = self.create_binary_package(
            source_package=depends,
            deb_fields={"Depends": "hello"},
            add_to_collection=self.sid,
        )
        binary_item = self.sid.child_items.get(artifact=binary_artifact)
        source_artifact = self.source_artifact
        architectures = ("amd64", "i386")

        root = self.playground.create_workflow(mark_running=True)
        reference = self.orchestrate(
            source_artifact,
            architectures,
            extra_data={
                "prefix": "reference-qa-result|",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
            },
            parent=root,
            pipeline_task_name="referencepipeline",
        )
        assert reference.parent is not None
        new = self.orchestrate(
            source_artifact,
            architectures,
            extra_data={
                "reference_prefix": "reference-qa-result|",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "enable_regression_tracking": True,
            },
            parent=root,
        )

        reference_child = reference.children.get()
        self.assertEqual(reference_child.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(reference_child.task_type, TaskTypes.WORKFLOW)
        self.assertEqual(reference_child.task_name, "autopkgtest")
        self.assertEqual(
            reference_child.task_data,
            {
                "prefix": "reference-qa-result|depends_1.0|",
                "reference_prefix": "depends_1.0|",
                "source_artifact": (
                    f"sid@{CollectionCategory.SUITE}/source-version:depends_1.0"
                ),
                "binary_artifacts": [
                    f"sid@{CollectionCategory.SUITE}/"
                    f"binary-version:depends_1.0_all"
                ],
                "context_artifacts": [
                    f"internal@collections/name:build-{architecture}"
                    for architecture in ("all", *architectures)
                ],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "enable_regression_tracking": False,
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
                "vendor": "debian",
                "codename": "sid",
                "backend": BackendType.UNSHARE,
                "architectures": [],
                "arch_all_build_architecture": "amd64",
                "debug_level": 0,
                "extra_repositories": None,
            },
        )

        reference_grandchild = reference_child.children.get(
            task_type=TaskTypes.WORKER, task_name="autopkgtest"
        )
        self.assertEqual(
            reference_grandchild.status, WorkRequest.Statuses.BLOCKED
        )
        self.assertEqual(reference_grandchild.task_type, TaskTypes.WORKER)
        self.assertEqual(reference_grandchild.task_name, "autopkgtest")
        self.assertEqual(
            reference_grandchild.task_data,
            {
                "input": {
                    "source_artifact": (
                        f"sid@{CollectionCategory.SUITE}/"
                        f"source-version:depends_1.0"
                    ),
                    "binary_artifacts": [
                        f"sid@{CollectionCategory.SUITE}/"
                        f"binary-version:depends_1.0_all"
                    ],
                    "context_artifacts": [
                        f"internal@collections/name:build-{architecture}"
                        for architecture in ("all", "amd64")
                    ],
                },
                "build_architecture": "amd64",
                "environment": "debian/match:codename=sid",
                "backend": BackendType.UNSHARE,
                "include_tests": [],
                "exclude_tests": [],
                "debug_level": 0,
                "extra_environment": {},
                "extra_repositories": None,
                "pin_packages": {},
                "needs_internet": AutopkgtestNeedsInternet.RUN,
                "fail_on": {},
                "timeout": None,
            },
        )
        qa_result_action = ActionUpdateCollectionWithArtifacts(
            collection=f"sid@{CollectionCategory.QA_RESULTS}",
            variables={
                "package": "depends",
                "version": "1.0",
                "architecture": "amd64",
                "timestamp": int(binary_item.created_at.timestamp()),
                "work_request_id": reference_grandchild.id,
            },
            artifact_filters={"category": ArtifactCategory.AUTOPKGTEST},
        )
        self.assert_work_request_event_reactions(
            reference_grandchild,
            on_assignment=[
                ActionSkipIfLookupResultChanged(
                    lookup=(
                        f"sid@{CollectionCategory.QA_RESULTS}/"
                        f"latest:autopkgtest_depends_amd64"
                    ),
                    promise_name=(
                        "reference-qa-result|depends_1.0|autopkgtest-amd64"
                    ),
                    collection_item_id=None,
                )
            ],
            on_failure=[qa_result_action],
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    collection="internal@collections",
                    name_template=(
                        "reference-qa-result|depends_1.0|autopkgtest-amd64"
                    ),
                    artifact_filters={"category": ArtifactCategory.AUTOPKGTEST},
                ),
                qa_result_action,
            ],
        )
        self.assertQuerySetEqual(
            reference_grandchild.dependencies.all(),
            # https://github.com/typeddjango/django-stubs/pull/3087
            reference.parent.children.get(  # type: ignore[arg-type]
                task_type=TaskTypes.WORKFLOW, task_name="sbuild"
            ).children.filter(
                task_data__build_architecture__in={"all", "amd64"}
            ),
        )
        self.assertEqual(
            reference_grandchild.workflow_data_json,
            {
                "allow_failure": True,
                "display_name": "autopkgtest amd64",
                "step": "autopkgtest-amd64",
            },
        )

        new_child = new.children.get()
        self.assertEqual(new_child.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(new_child.task_type, TaskTypes.WORKFLOW)
        self.assertEqual(new_child.task_name, "autopkgtest")
        self.assertEqual(
            new_child.task_data,
            {
                "prefix": "depends_1.0|",
                "reference_prefix": "reference-qa-result|depends_1.0|",
                "source_artifact": (
                    f"sid@{CollectionCategory.SUITE}/source-version:depends_1.0"
                ),
                "binary_artifacts": [
                    f"sid@{CollectionCategory.SUITE}/"
                    f"binary-version:depends_1.0_all"
                ],
                "context_artifacts": [
                    f"internal@collections/name:build-{architecture}"
                    for architecture in ("all", *architectures)
                ],
                "qa_suite": f"sid@{CollectionCategory.SUITE}",
                "reference_qa_results": f"sid@{CollectionCategory.QA_RESULTS}",
                "enable_regression_tracking": True,
                "update_qa_results": RegressionTrackingUpdateQAResults.NO,
                "vendor": "debian",
                "codename": "sid",
                "backend": BackendType.UNSHARE,
                "architectures": [],
                "arch_all_build_architecture": "amd64",
                "debug_level": 0,
                "extra_repositories": None,
            },
        )

        new_grandchild = new_child.children.get(
            task_type=TaskTypes.WORKER, task_name="autopkgtest"
        )
        self.assertEqual(new_grandchild.status, WorkRequest.Statuses.BLOCKED)
        self.assertEqual(new_grandchild.task_type, TaskTypes.WORKER)
        self.assertEqual(new_grandchild.task_name, "autopkgtest")
        self.assertEqual(
            new_grandchild.task_data,
            {
                "input": {
                    "source_artifact": (
                        f"sid@{CollectionCategory.SUITE}/"
                        f"source-version:depends_1.0"
                    ),
                    "binary_artifacts": [
                        f"sid@{CollectionCategory.SUITE}/"
                        f"binary-version:depends_1.0_all"
                    ],
                    "context_artifacts": [
                        f"internal@collections/name:build-{architecture}"
                        for architecture in ("all", "amd64")
                    ],
                },
                "build_architecture": "amd64",
                "environment": "debian/match:codename=sid",
                "backend": BackendType.UNSHARE,
                "include_tests": [],
                "exclude_tests": [],
                "debug_level": 0,
                "extra_environment": {},
                "extra_repositories": None,
                "pin_packages": {},
                "needs_internet": AutopkgtestNeedsInternet.RUN,
                "fail_on": {},
                "timeout": None,
            },
        )
        self.assert_work_request_event_reactions(
            new_grandchild,
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    collection="internal@collections",
                    name_template="depends_1.0|autopkgtest-amd64",
                    artifact_filters={"category": ArtifactCategory.AUTOPKGTEST},
                )
            ],
        )
        self.assertQuerySetEqual(
            new_grandchild.dependencies.all(),
            # https://github.com/typeddjango/django-stubs/pull/3087
            reference.parent.children.get(  # type: ignore[arg-type]
                task_type=TaskTypes.WORKFLOW, task_name="sbuild"
            ).children.filter(
                task_data__build_architecture__in={"all", "amd64"}
            ),
        )
        self.assertEqual(
            new_grandchild.workflow_data_json,
            {"display_name": "autopkgtest amd64", "step": "autopkgtest-amd64"},
        )

        # Population is idempotent.
        self.get_workflow(reference).populate()
        reference_children = list(WorkRequest.objects.filter(parent=reference))
        self.assertEqual(len(reference_children), 1)
        self.get_workflow(new).populate()
        new_children = list(WorkRequest.objects.filter(parent=new))
        self.assertEqual(len(new_children), 1)

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        context_artifact = (
            self.playground.create_minimal_binary_package_artifact()
        )
        self.task_data["context_artifacts"] = [context_artifact.id]
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            ReverseDependenciesAutopkgtestWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello_1.0-1",
                source_artifact_id=self.source_artifact.id,
                binary_artifacts_ids=[self.binary_artifact.id],
                context_artifacts_ids=[context_artifact.id],
                qa_suite_id=self.sid.id,
            ),
        )

    def test_get_input_artifacts(self) -> None:
        binary_artifact_1 = (
            self.playground.create_minimal_binary_package_artifact(
                package="hello"
            )
        )
        binary_artifact_2 = (
            self.playground.create_minimal_binary_package_artifact(
                package="bye"
            )
        )
        context_artifact_1 = (
            self.playground.create_minimal_binary_package_artifact(
                package="python3"
            )
        )
        context_artifact_2 = (
            self.playground.create_minimal_binary_package_artifact(
                package="python3-django"
            )
        )
        self.task_data["binary_artifacts"] = [
            binary_artifact_1.id,
            binary_artifact_2.id,
        ]
        self.task_data["context_artifacts"] = [
            context_artifact_1.id,
            context_artifact_2.id,
        ]
        workflow = self.task
        workflow.dynamic_data = workflow.compute_dynamic_data()
        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=self.source_artifact.id,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[binary_artifact_1.id, binary_artifact_2.id],
                ),
                InputArtifactMultiple.from_ids(
                    lookup=workflow.data.context_artifacts,
                    label="context_artifacts",
                    artifact_ids=[context_artifact_1.id, context_artifact_2.id],
                ),
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Table column filtering."""

import abc
from collections.abc import Collection
from functools import cached_property
from typing import Any, NamedTuple, TYPE_CHECKING, override

from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model, Q, QuerySet
from django.forms import Form, modelform_factory
from django.forms.boundfield import BoundField
from django.http import QueryDict
from django.template.context import Context
from django.utils.html import format_html
from django.utils.safestring import SafeString

from debusine.web.forms import ModelFormBase
from debusine.web.views.base import TemplateWidget, Widget
from debusine.web.views.table.declarative import Attribute

if TYPE_CHECKING:
    from debusine.web.views.table.table import Table


class FilterWidget(Widget):
    """Base class for menu entry widgets."""

    def __init__(self, filter_: "BoundFilter[Any]") -> None:
        """Store a reference to the bound filter."""
        self.filter = filter_


class MenuEntry(FilterWidget):
    """Standard menu entry, activating the filter form."""

    @override
    def render(self, context: Context) -> str:  # noqa: ARG002, U100
        """Render the filter as widget."""
        return format_html(
            "<a class='dropdown-item debusine-dropdown-remote'"
            " aria-expanded='false'"
            " data-bs-target='#{dom_id_form_dropdown}'>"
            "{label}</a>",
            dom_id_form_dropdown=self.filter.dom_id_form_dropdown,
            label=self.filter.label,
        )


class MenuEntryToggle(FilterWidget):
    """Menu entry that toggles the filter value."""

    @override
    def render(self, context: Context) -> str:  # noqa: ARG002, U100
        """Render the filter as widget."""
        if self.filter.value:
            url = self.filter.qs_remove
        else:
            url = self.filter.qs_with_value(True)
        return format_html(
            "<a class='dropdown-item{active}'"
            " href='?{url}' {aria_current}>"
            "{label}</a>",
            url=url,
            label=self.filter.label,
            active=" active" if self.filter.value else "",
            aria_current=(
                SafeString(" aria-current='true'") if self.filter.value else ""
            ),
        )


class MainEntry(FilterWidget, TemplateWidget):
    """Render the element and optional dropdown form for a filter."""

    @override
    def get_context_data(self) -> dict[str, Any]:
        ctx = super().get_context_data()
        ctx["filter"] = self.filter
        ctx["table"] = self.filter.filters.table
        return ctx


class MainEntryForm(MainEntry):
    """Render input forms."""

    template_name = "web/_table_filter_form.html"
    form_template_name: str

    @override
    def get_context_data(self) -> dict[str, Any]:
        ctx = super().get_context_data()
        ctx["form_template_name"] = self.form_template_name
        return ctx


class MainEntryText(MainEntryForm):
    """Render the input field for a FilterText."""

    form_template_name = "web/_table_filter_text.html"


class MainEntrySelectOne(MainEntryForm):
    """Render the input field for a FilterSelectOne."""

    form_template_name = "web/_table_filter_select_one.html"


class MainEntryField(MainEntryForm):
    """Render the input field for a FilterField."""

    form_template_name = "web/_table_filter_field.html"


class MainEntryToggle(MainEntry):
    """Render the input field for a FilterToggle."""

    template_name = "web/_table_filter_toggle.html"


class BoundFilter[M: Model]:
    """A Filter bound to a table."""

    dom_id: str
    dom_id_form_dropdown: str

    def __init__(
        self,
        filter_: "Filter",
        filters: "Filters[M]",
    ) -> None:
        """Build from filter and table filters definition."""
        self.filter = filter_
        self.filters = filters
        #: ID for this element in the DOM
        self.dom_id = self.filters.table.dom_id + f"filter-{self.name}"
        self.dom_id_form_dropdown = self.dom_id + "-form-dropdown"

    @cached_property
    def filter_prefix(self) -> str:
        """Compute the filter prefix."""
        return f"{self.filters.filter_prefix}-{self.name}"

    @property
    def name(self) -> str:
        """Return the filter name."""
        return self.filter.name

    @property
    def label(self) -> str:
        """Return the filter label."""
        return self.filter.label

    @property
    def icon(self) -> str | None:
        """Return the filter icon, if defined."""
        return self.filter.icon

    @cached_property
    def value(self) -> Any:
        """Return the current value for the filter."""
        return self.filters.table.request.GET.get(self.filter_prefix) or ""

    @cached_property
    def active(self) -> bool:
        """Check if the filter is active."""
        return bool(self.value)

    def _format_value(self) -> str:
        """Return a pretty printed value."""
        match value := self.value:
            case None:
                return ""
            case str():
                return value
            case Collection():
                return ", ".join(value)
            case _:
                return str(value)

    @cached_property
    def value_formatted(self) -> str:
        """Return a pretty printed value."""
        return self._format_value()

    def value_for_querystring(self, value: Any) -> list[str]:
        """
        Convert the current value to a list of strings for use in a querystring.

        Return an empty list if the given value means the filter is disabled.
        """
        match value:
            case True:
                return ["1"]
            case None | False:
                return []
            case str():
                return [value]
            case Collection():
                return [str(v) for v in value]
            case _:
                return [str(value)]

    def qs_with_value(self, value: Any) -> str:
        """Get a query string setting the field value to a given value."""
        base_args = QueryDict(mutable=True)
        for key, values in self.filters.table.request.GET.lists():
            if key.startswith(self.filter_prefix):
                continue
            base_args.setlist(key, values)
        if v := self.value_for_querystring(value):
            base_args.setlist(self.filter_prefix, v)
        return base_args.urlencode()

    @cached_property
    def qs_remove(self) -> str:
        """Query string that can be used to remove this filter."""
        base_args = QueryDict(mutable=True)
        for key, values in self.filters.table.request.GET.lists():
            if key.startswith(self.filter_prefix):
                continue
            base_args.setlist(key, values)
        return base_args.urlencode()

    @cached_property
    def base_filter_args(self) -> list[tuple[str, str]]:
        """Form arguments for all aspects of this table except this filter."""
        base_arglist: list[tuple[str, str]] = []
        for key, values in self.filters.table.request.GET.lists():
            if key.startswith(self.filter_prefix):
                continue
            base_arglist += [(key, value) for value in values]
        return base_arglist

    def add_option(
        self,
        name: str,  # noqa: ARG002, U100
        label: str,  # noqa: ARG002, U100
        query: Q,  # noqa: ARG002, U100
        icon: str | None = None,  # noqa: ARG002, U100
    ) -> None:
        """Add an option to the selection list."""
        raise ImproperlyConfigured(
            "add_option called on a field that does not support options."
        )

    def filter_queryset(self, queryset: QuerySet[M, M]) -> QuerySet[M, M]:
        """Filter the queryset."""
        return self.filter.filter_queryset(queryset, self.value)

    @cached_property
    def main_entry(self) -> FilterWidget:
        """Return the label to use to show the active filter."""
        return self.filter.main_entry(self)

    @cached_property
    def menu_entry(self) -> FilterWidget:
        """Return the widget to render our entry in the filter menu."""
        return self.filter.menu_entry(self)


class Filter(Attribute, abc.ABC):
    """A way of filtering a table."""

    label: str
    icon: str | None
    bound_class: type[BoundFilter[Any]] = BoundFilter[Any]
    menu_entry: type[FilterWidget] = MenuEntry
    main_entry: type[FilterWidget]

    def __init__(
        self,
        label: str,
        *,
        name: str | None = None,
        icon: str | None = None,
    ) -> None:
        """Store filter definition."""
        super().__init__(name=name)
        self.label = label
        self.icon = icon

    @abc.abstractmethod
    def filter_queryset[M: Model](
        self, queryset: QuerySet[M, M], value: Any
    ) -> QuerySet[M, M]:
        """Filter the queryset using the given filter value."""


class FilterText(Filter):
    """A filter with a text field."""

    main_entry = MainEntryText
    placeholder: str | None

    def __init__(
        self,
        label: str,
        *,
        q_field: str | None = None,
        placeholder: str | None = None,
        icon: str | None = None,
    ) -> None:
        """
        Store filter definition.

        :param q_field: field name to use for matching in Q expressions.
                        Defaults to ``"{self.name}__contains"``
        :param placeholder: Placeholder string to show in the text field
        """
        super().__init__(label, icon=icon)
        self._q_field = q_field
        self.placeholder = placeholder

    @cached_property
    def q_field(self) -> str:
        """Return the field name used for matching in Q expressions."""
        return self._q_field or f"{self.name}__contains"

    @override
    def filter_queryset[M: Model](
        self, queryset: QuerySet[M, M], value: Any
    ) -> QuerySet[M, M]:
        return queryset.filter(**{self.q_field: value})


class FilterToggle(Filter):
    """A filter that can be toggled on or off."""

    menu_entry: type[FilterWidget] = MenuEntryToggle
    main_entry: type[FilterWidget] = MainEntryToggle

    def __init__(
        self,
        label: str,
        *,
        q: Q,
        icon: str | None = None,
    ) -> None:
        """
        Store filter definition.

        :param q: Q expression to use for filtering
        """
        super().__init__(label, icon=icon)
        self.q = q

    @override
    def filter_queryset[M: Model](
        self, queryset: QuerySet[M, M], value: Any
    ) -> QuerySet[M, M]:
        if not value:
            return queryset
        return queryset.filter(self.q)


class SelectOption(NamedTuple):
    """One option in a select/multiselect filter."""

    name: str
    label: str
    query: Q
    icon: str | None = None


class BoundFilterSelectOne[M: Model](BoundFilter[M]):
    """Filter specialization handling a set of selectable values."""

    def __init__(
        self,
        filter_: "FilterSelectOne",
        filters: "Filters[M]",
    ) -> None:
        """Build from filter and table filters definition."""
        super().__init__(filter_, filters)
        self.options: dict[str, SelectOption] = {}
        for option in filter_.options:
            self.add(option)

    def add(self, option: SelectOption) -> None:
        """Add an option to the selection list."""
        self.options[option.name] = option

    @override
    def add_option(
        self, name: str, label: str, query: Q, icon: str | None = None
    ) -> None:
        self.add(SelectOption(name=name, label=label, query=query, icon=icon))

    @override
    def filter_queryset(self, queryset: QuerySet[M, M]) -> QuerySet[M, M]:
        if option := self.options.get(self.value):
            return queryset.filter(option.query)
        else:
            return queryset


class FilterSelectOne(Filter):
    """Filter selecting one item among a finite set."""

    bound_class = BoundFilterSelectOne
    main_entry = MainEntrySelectOne

    #: Options to select from
    options: Collection[SelectOption]

    def __init__(
        self,
        label: str,
        *,
        icon: str | None = None,
        options: Collection[SelectOption] = (),
    ) -> None:
        """
        Store filter definition.

        :param options: Collection of ``SelectOption`` elements describing the
                        options the user can choose from. Defaults to empty.
        """
        super().__init__(label, icon=icon)
        self.options = options

    @override
    def filter_queryset[M: Model](
        self,
        queryset: QuerySet[M, M],  # noqa: ARG002, U100
        value: Any,  # noqa: ARG002, U100
    ) -> QuerySet[M, M]:
        raise ImproperlyConfigured(
            "FilterSelectOne should have bound_class"
            " set to a BoundFilterSelectOne"
        )


class BoundFilterField[M: Model](BoundFilter[M]):
    """Filter specialization using a Django Form field."""

    @override
    @cached_property
    def value(self) -> Any:
        return self.filters.form.cleaned_data.get(self.name)

    @override
    def _format_value(self) -> str:
        if self.value is None:
            return ""

        # If we are using a choice field, get its display value
        if choices := getattr(self.form_field.field, "choices", None):
            values: Collection[str] = (
                [self.value] if isinstance(self.value, str) else self.value
            )
            labels: list[str] = []
            for k, v in choices:
                if isinstance(v, (list, tuple)):
                    # This is an optgroup, so look inside the group for options
                    for k2, v2 in v:
                        if k2 in values:
                            labels.append(v2)
                else:
                    if k in values:
                        labels.append(v)
            if labels:
                return ", ".join(labels)

        return super()._format_value()

    @cached_property
    def form_field(self) -> BoundField:
        """Return the bound Django form field for this filter."""
        try:
            return self.filters.form[self.name]
        except KeyError:
            available = repr(sorted(self.filters.form.fields.keys()))
            raise ImproperlyConfigured(
                f"{self.name!r} not found"
                f" in table filter form {self.filters.form.__class__.__name__}."
                f" Available fields: {available}"
            )

    @override
    @cached_property
    def filter_prefix(self) -> str:
        return self.form_field.html_name

    @override
    @property
    def label(self) -> str:
        return self.filter.label or str(self.form_field.label)


class FilterField(Filter):
    """Table filter using a Django form field."""

    bound_class = BoundFilterField
    main_entry = MainEntryField

    def __init__(
        self,
        label: str = "",
        icon: str | None = None,
        q_field: str | None = None,
    ) -> None:
        """Store filter definition."""
        super().__init__(label, icon=icon)
        self._q_field = q_field

    @cached_property
    def q_field(self) -> str:
        """Return the field name used for matching in Q expressions."""
        return self._q_field or f"{self.name}__contains"

    @override
    def filter_queryset[M: Model](
        self, queryset: QuerySet[M, M], value: Any
    ) -> QuerySet[M, M]:
        return queryset.filter(**{self.q_field: value})


class Filters[M: Model](TemplateWidget):
    """Filter definitions for a table."""

    template_name = "web/_table_filters.html"

    def __init__(self, table: "Table[M]") -> None:
        """Instantiate for the given table."""
        self.table = table

        #: Prefix for filter query string arguments for this table
        self.filter_prefix: str = f"{self.table.prefix}filter"

        #: ID for this element in the DOM
        self.dom_id = self.table.dom_id + "filters"

        #: ID for the filters menu in the DOM
        self.dom_id_filters_menu = self.dom_id + "-menu"

        #: Filters configured for the table
        self.options: dict[str, BoundFilter[M]] = {}

        # Set up query string related members
        query = QueryDict(mutable=True)
        query_items: list[tuple[str, str]] = []
        for key, values in self.table.request.GET.lists():
            if key.startswith(self.filter_prefix):
                continue
            query.setlist(key, values)
            query_items += [(key, value) for value in values]

        #: Query string clearing all filters for this table
        self.qs_clear_filters = query.urlencode()
        self.non_filter_args = query_items

    def __bool__(self) -> bool:
        """Return True if there are filters defined."""
        return bool(self.options)

    @cached_property
    def available(self) -> Collection[BoundFilter[M]]:
        """Return all available filters."""
        return self.options.values()

    def __getitem__(self, name: str) -> BoundFilter[M]:
        """Get a filter by name."""
        return self.options[name]

    @cached_property
    def active(self) -> Collection[BoundFilter[M]]:
        """Return all active filters."""
        return [v for v in self.available if v.active]

    @cached_property
    def inactive(self) -> Collection[BoundFilter[M]]:
        """Return all active filters."""
        return [v for v in self.available if not v.active]

    @cached_property
    def form(self) -> Form | ModelFormBase[M]:
        """
        Form to use for table filter input.

        If none is specified, one is automatically inferred using django
        modelforms.
        """
        form_class: type[Form] | type[ModelFormBase[M]]
        if self.table.filter_form_class:
            form_class = self.table.filter_form_class
        else:

            class Base(ModelFormBase[M]):
                """ModelForm base that allows empty and existing values."""

                def __init__(self, *args: Any, **kwargs: Any) -> None:
                    """Override ModelForm defaults."""
                    kwargs["empty_permitted"] = True
                    kwargs["use_required_attribute"] = False
                    super().__init__(*args, **kwargs)

                @override
                def validate_unique(self) -> None:
                    """Disable unique validation."""

            # Infer the list of fields from the list of FilterField entries
            field_names: list[str] = []
            for name, flt in self.options.items():
                if not isinstance(flt.filter, FilterField):
                    continue
                field_names.append(name)
            form_class = modelform_factory(
                self.table.model,
                form=Base,
                fields=field_names,
            )
        form = form_class(self.table.request.GET, prefix=self.filter_prefix)
        form.is_valid()
        return form

    def add(self, filter_: Filter) -> None:
        """Add a new filter."""
        self.options[filter_.name] = filter_.bound_class(filter_, self)

    def filter_queryset(self, queryset: QuerySet[M, M]) -> QuerySet[M, M]:
        """Filter the queryset using the active filters."""
        for f in self.options.values():
            if f.active:
                queryset = f.filter_queryset(queryset)
        return queryset

    @override
    def get_context_data(self) -> dict[str, Any]:
        ctx = super().get_context_data()
        ctx["filters"] = self
        return ctx



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""URL manipulation utilities."""

import re
from collections.abc import Sequence
from typing import Literal, Self, overload, override

# (connect, read) timeout for external HTTP requests
DEFAULT_REQUEST_TIMEOUT = (5, 30)

re_url_path_sep = re.compile(r"/+")


class URLPath(Sequence[str]):
    """
    Store and manipulate the path component of a URL.

    It behaves as a list of split path components.
    """

    #: True if the path refers to a file, False if it refers to a directory
    is_file: bool

    def __init__(
        self,
        path: str = "",
        is_file: bool | Literal["autodetect"] = "autodetect",
    ) -> None:
        """
        Construct a URLPath from a path string.

        :param path: the path component of a parsed URL
        :param is_file: if the path refers to a file or a directory. If unset,
          it will be autodetected by checking if path ends with a slash

        Multiple consecutive slashes in the path are treated like a single
        slash instead of empty path components.

        Leading slashes in path are ignored.
        """
        stripped_path = path.strip("/")
        self.path = (
            re_url_path_sep.split(stripped_path) if stripped_path else []
        )
        if is_file == "autodetect":
            self.is_file = bool(stripped_path) and not path.endswith("/")
        else:
            self.is_file = is_file

    @override
    def __str__(self) -> str:
        """
        Return the path as a string.

        Paths referring to a directory will end with a trailing slash.
        """
        res = "/".join(self.path)
        if not self.is_file:
            res += "/"
        return res

    @override
    def __len__(self) -> int:
        """Return the number of path components."""
        return len(self.path)

    @overload
    def __getitem__(self, idx: int, /) -> str: ...  # noqa: D105

    @overload
    def __getitem__(self, idx: slice, /) -> Sequence[str]: ...  # noqa: D105

    @override
    def __getitem__(self, idx: int | slice, /) -> str | Sequence[str]:
        """Return a path component by its list index."""
        return self.path[idx]

    def __add__(self, value: "URLPath | str") -> "URLPath":
        """
        Append two URLPath objects.

        If the first URLPath is a file, the file component is removed.

        Adding a string will automatically convert it to a URLPath, with
        is_file autodetected from the presence of a trailing slash.
        """
        match value:
            case URLPath():
                pass
            case str():
                value = URLPath(value)
            case _:
                return NotImplemented
        res = URLPath()
        if self.is_file:
            res.path += self.path[:-1]
        else:
            res.path += self.path
        res.path += value.path
        res.is_file = value.is_file
        return res

    def __iadd__(self, value: "URLPath | str") -> Self:
        """
        In-place append.

        If this URLPath is a file, the file component is removed.

        Adding a string will automatically convert it to a URLPath, with
        is_file autodetected from the presence of a trailing slash.
        """
        match value:
            case URLPath():
                pass
            case str():
                value = URLPath(value)
            case _:
                return NotImplemented
        if self.is_file:
            self.pop()
        self.path += value.path
        self.is_file = value.is_file
        return self

    def pop(self) -> None:
        """Remove the last element from the path."""
        self.is_file = False
        if self.path:
            self.path.pop()


def parse_timeout(raw_timeout: str) -> int | tuple[int, int]:
    """Parse a string into an integer or a tuple of two integers."""
    try:
        if "," in raw_timeout:
            as_ints = tuple(int(x.strip()) for x in raw_timeout.split(",", 1))
            return (as_ints[0], as_ints[1])
        return int(raw_timeout)
    except (ValueError, IndexError):
        raise ValueError(f"Invalid timeout value: {raw_timeout}")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Context for database and permission operations.

This is currently based on ContextVar instead of asgiref.local.Local, because
the behaviour of Local is inconsistent across asgiref versions, and probably
buggy in asgiref 3.7+ (see https://github.com/django/asgiref/issues/473).

Ideally, this should eventually aligned with using the same
thread-local/task-local infrastructure that Django uses.

Note that when using threads, if a new thread is started, it will see the
application context reset to empty values: threads that need application
context values need to implement a way to inherit the caller's values.

Note that similar but more complicated gotchas would apply when
`django.urls.reverse` tries to check if `request.urlconf` was previously set to
a non-default value: threads and tasks that need reverse url resolution need to
be mindful of the inconsistencies in behaviour of the underlying
`asgiref.local.Local` implementation, and of asgiref issue #473.

See also https://code.djangoproject.com/ticket/35807
"""

from collections.abc import Generator
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Optional, TYPE_CHECKING

from debusine.db.permissioncontext import PermissionContext, PermissionUser

if TYPE_CHECKING:
    from debusine.db.models import Scope, Token, Workspace
    from debusine.db.models.lookups import LookupCache
    from debusine.db.models.scopes import ScopeRoles
    from debusine.db.models.workspaces import WorkspaceRoles


class ContextConsistencyError(Exception):
    """Raised if an inconsistency is found when setting application context."""


class Context:
    """
    Storage for Debusine application context.

    Setting the application context needs to be done in a specific order:

    1. ``context.set_scope()``
    2. ``context.set_permission_context()``
    3. ``workspace.set_current()``

    This reflects the order in which information is available in the middleware
    chain, where scope is needed first in order to activate URL resolution,
    which is a prerequisite of the auth layer which will then populate the user.
    """

    # https://github.com/django/asgiref/issues/473

    # Use slots to catch typos in functions that set context variables
    __slots__ = [
        "_lookup_cache",
        "_permission_context",
        "_scope",
        "_scope_roles",
        "_workspace",
        "_workspace_roles",
    ]

    def __init__(self) -> None:
        """Initialize the default values."""
        super().__init__()

        self._lookup_cache: ContextVar[LookupCache | None] = ContextVar(
            "lookup_cache", default=None
        )

        self._permission_context: ContextVar[PermissionContext | None] = (
            ContextVar("permission_context", default=None)
        )

        self._scope: ContextVar[Scope | None] = ContextVar(
            "scope", default=None
        )

        self._scope_roles: ContextVar[frozenset[ScopeRoles] | None] = (
            ContextVar("scope_roles", default=None)
        )

        self._workspace: ContextVar[Workspace | None] = ContextVar(
            "workspace", default=None
        )
        self._workspace_roles: ContextVar[frozenset[WorkspaceRoles] | None] = (
            ContextVar("workspace_roles", default=None)
        )

    @property
    def permission_checks_disabled(self) -> bool:
        """
        Return True if permission checks should always succeed.

        This is used to skip permission checks in situations like management
        commands or test fixture setup.
        """
        if (pc := self._permission_context.get()) is None:
            return False
        return pc.permission_checks_disabled

    @contextmanager
    def disable_permission_checks(self) -> Generator[None]:
        """Set admin mode for the duration of the context manager."""
        if self._permission_context.get() is None:
            orig = self._permission_context.set(
                PermissionContext.for_ui_anonymous().with_permission_checks_disabled()
            )
        else:
            orig = self._permission_context.set(
                self.pc.with_permission_checks_disabled()
            )
        try:
            yield
        finally:
            self._permission_context.reset(orig)

    @property
    def pc(self) -> PermissionContext:
        """Return the current permission context."""
        res = self._permission_context.get()
        if res is None:
            raise ContextConsistencyError(
                "permission context has not yet been set"
            )
        return res

    def pc_is(self, pc: PermissionContext) -> bool:
        """Return True if the current permission context, if set, matches pc."""
        return self._permission_context.get() == pc

    def set_permission_context_scoped(
        self, scope: "Scope", pc: PermissionContext
    ) -> None:
        """Set the permission context, for scoped operations."""
        from debusine.db.models.scopes import ScopeRoles

        if (old_scope := self.scope) is not None:
            raise ContextConsistencyError(
                f"Scope was already set to {old_scope}"
            )

        if (old := self._permission_context.get()) is not None:
            raise ContextConsistencyError(
                f"Permission context was already set to {old}"
            )

        self._scope.set(scope)
        self._scope_roles.set(None)
        self._permission_context.set(pc)
        self._workspace.set(None)
        self._workspace_roles.set(None)

        # TODO: Scope visibility checks should happen here. Currently scopes
        # are always visible, so the only thing needed is this comment as a
        # placeholder

        self._scope_roles.set(
            ScopeRoles.from_iterable(scope.get_group_roles(pc))
        )

    def set_permission_context_scopeless(self, pc: PermissionContext) -> None:
        """Set the permission context, for unscoped operations."""
        if self.scope is not None:
            raise ContextConsistencyError(
                f"Scope was already set to {self.scope}"
            )

        if (old := self._permission_context.get()) is not None:
            raise ContextConsistencyError(
                f"Permission context was already set to {old}"
            )

        self._permission_context.set(pc)
        self._scope_roles.set(None)
        self._workspace.set(None)
        self._workspace_roles.set(None)

    @property
    def scope(self) -> Optional["Scope"]:
        """Get the current scope."""
        return self._scope.get()

    def require_scope(self) -> "Scope":
        """
        Get the current scope.

        :raises ContextConsistencyError: if the scope is not set
        """
        if (scope := self.scope) is None:
            raise ContextConsistencyError("scope is not set in context")
        return scope

    @property
    def user(self) -> PermissionUser:
        """Return the current user."""
        return self.pc.user

    @property
    def user_token(self) -> Optional["Token"]:
        """Get the current user token."""
        return self.pc.user_token

    @property
    def external_user_token(self) -> Optional["Token"]:
        """Get the current external user token."""
        return self.pc.external_user_token

    @property
    def workspace(self) -> Optional["Workspace"]:
        """Get the current workspace."""
        return self._workspace.get()

    def require_workspace(self) -> "Workspace":
        """
        Get the current workspace.

        :raises ContextConsistencyError: if the workspace is not set
        """
        if (workspace := self.workspace) is None:
            raise ContextConsistencyError("workspace is not set in context")
        return workspace

    @property
    def scope_roles(self) -> frozenset["ScopeRoles"]:
        """Get the roles the current user has on the current Scope."""
        if (roles := self._scope_roles.get()) is not None:
            return roles
        raise ContextConsistencyError("scope has not been set")

    @property
    def workspace_roles(self) -> frozenset["WorkspaceRoles"]:
        """Get the roles the current user has on the current Workspace."""
        if (roles := self._workspace_roles.get()) is not None:
            return roles
        raise ContextConsistencyError("workspace is not set")

    @property
    def lookup_cache(self) -> "LookupCache":
        """Get the current context's lookup cache."""
        # Avoid circular import.
        from debusine.db.models.lookups import LookupCache

        if (lookup_cache := self._lookup_cache.get()) is not None:
            return lookup_cache
        self._lookup_cache.set(lookup_cache := LookupCache())
        return lookup_cache

    def clear_lookup_cache(self) -> None:
        """Clear the current context's lookup cache."""
        self._lookup_cache.set(None)

    def reset(self) -> None:
        """Reset the application context to default values."""
        self._permission_context.set(None)
        self._scope.set(None)
        self._scope_roles.set(None)
        self._workspace.set(None)
        self._workspace_roles.set(None)
        self._lookup_cache.set(None)

    @contextmanager
    def local(self) -> Generator[None]:
        """Restore application context when the context manager ends."""
        orig_pc = self._permission_context.get()
        orig_scope = self.scope
        orig_scope_roles = self._scope_roles.get()
        orig_workspace = self.workspace
        orig_workspace_roles = self._workspace_roles.get()
        orig_lookup_cache = self.lookup_cache
        self._lookup_cache.set(None)
        try:
            yield
        finally:
            self._permission_context.set(orig_pc)
            self._scope.set(orig_scope)
            self._scope_roles.set(orig_scope_roles)
            self._workspace.set(orig_workspace)
            self._workspace_roles.set(orig_workspace_roles)
            self._lookup_cache.set(orig_lookup_cache)

    @contextmanager
    def management_command(self) -> Generator[None]:
        """
        Set a context for management commands.

        This only sets the context user to the system user, so that group
        management commands can create an appropriate audit log.
        """
        with self.local():
            self.reset()
            pc = PermissionContext.for_maintenance()
            self._permission_context.set(pc)
            yield


context = Context()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command signing_worker."""

from unittest import mock
from unittest.mock import AsyncMock

from django.core.management import call_command
from django.test import TestCase

from debusine.tasks.models import WorkerType


class SigningWorkerCommandTests(TestCase):
    """Tests for the signing_worker command."""

    def test_no_options(self) -> None:
        """Default to making a signing worker logging to stderr."""
        with (
            mock.patch(
                "debusine.signing.management.commands.signing_worker.Worker",
                autospec=True,
            ) as mock_worker,
            self.assertRaises(SystemExit),
        ):
            mock_worker().main = AsyncMock()

            call_command("signing_worker")

        mock_worker.assert_called_with(
            log_level=None, worker_type=WorkerType.SIGNING
        )

    def test_log_level(self) -> None:
        """The --log-level option is passed through to the worker."""
        with (
            mock.patch(
                "debusine.signing.management.commands.signing_worker.Worker",
                autospec=True,
            ) as mock_worker,
            self.assertRaises(SystemExit),
        ):
            mock_worker().main = AsyncMock()

            call_command("signing_worker", "--log-level", "DEBUG")

        mock_worker.assert_called_with(
            log_level="DEBUG", worker_type=WorkerType.SIGNING
        )

    def test_run_worker(self) -> None:
        """Run a signing worker."""
        with (
            mock.patch(
                "debusine.worker._worker.ConfigHandler", autospec=True
            ) as mock_config_handler,
            mock.patch(
                "debusine.worker.Worker.main", autospec=True
            ) as mock_main,
            self.assertRaises(SystemExit) as raised,
        ):
            mock_config_handler().log_level = None

            call_command("signing_worker")

        mock_main.assert_awaited_once()
        self.assertEqual(raised.exception.code, 0)
